"""Provides basis translators for SMEFT and and WET that can be used with the
`wcxf` Python package."""


from . import smeft
from . import wet
import wcxf


@wcxf.translator('SMEFT', 'Warsaw', 'Warsaw mass')
def warsaw_to_warsawmass(C, scale, parameters):
    return smeft.warsaw_to_warsawmass(C)


@wcxf.translator('SMEFT', 'Warsaw', 'Warsaw up')
def warsaw_to_warsaw_up(C, scale, parameters):
    return smeft.warsaw_to_warsaw_up(C)


@wcxf.translator('SMEFT', 'Warsaw up', 'Warsaw')
def warsaw_up_to_warsaw(C, scale, parameters):
    return smeft.warsaw_up_to_warsaw(C)


@wcxf.translator('WET', 'flavio', 'JMS')
def flavio_to_JMS(C, scale, parameters):
    return wet.flavio_to_JMS(C, scale, parameters)


@wcxf.translator('WET-4', 'flavio', 'JMS')
def flavio_to_JMS_wet4(C, scale, parameters):
    return wet.flavio_to_JMS(C, scale, parameters)


@wcxf.translator('WET-3', 'flavio', 'JMS')
def flavio_to_JMS_wet3(C, scale, parameters):
    return wet.flavio_to_JMS(C, scale, parameters)


@wcxf.translator('WET', 'JMS', 'flavio')
def JMS_to_flavio(C, scale, parameters):
    return wet.JMS_to_flavio(C, scale, parameters)


@wcxf.translator('WET-4', 'JMS', 'flavio')
def JMS_to_flavio_wet4(C, scale, parameters):
    return wet.JMS_to_flavio(C, scale, parameters)


@wcxf.translator('WET-3', 'JMS', 'flavio')
def JMS_to_flavio_wet3(C, scale, parameters):
    return wet.JMS_to_flavio(C, scale, parameters)


@wcxf.translator('WET', 'Bern', 'flavio')
def Bern_to_flavio(C, scale, parameters):
    return wet.Bern_to_flavio(C, scale, parameters)


@wcxf.translator('WET', 'flavio', 'Bern')
def flavio_to_Bern(C, scale, parameters):
    return wet.flavio_to_Bern(C, scale, parameters)


@wcxf.translator('WET-4', 'Bern', 'flavio')
def Bern_to_flavio_wet4(C, scale, parameters):
    return wet.Bern_to_flavio(C, scale, parameters)


@wcxf.translator('WET-4', 'flavio', 'Bern')
def flavio_to_Bern_wet4(C, scale, parameters):
    return wet.flavio_to_Bern(C, scale, parameters)


@wcxf.translator('WET-3', 'Bern', 'flavio')
def Bern_to_flavio_wet3(C, scale, parameters):
    return wet.Bern_to_flavio(C, scale, parameters)


@wcxf.translator('WET-3', 'flavio', 'Bern')
def flavio_to_Bern_wet3(C, scale, parameters):
    return wet.flavio_to_Bern(C, scale, parameters)


@wcxf.translator('WET', 'JMS', 'EOS')
def JMS_to_EOS(C, scale, parameters):
    return wet.JMS_to_EOS(C, scale, parameters)


@wcxf.translator('WET', 'JMS', 'Bern')
def JMS_to_Bern(C, scale, parameters):
    return wet.JMS_to_Bern(C, scale, parameters)


@wcxf.translator('WET-4', 'JMS', 'Bern')
def JMS_to_Bern_wet4(C, scale, parameters):
    return wet.JMS_to_Bern(C, scale, parameters)


@wcxf.translator('WET-3', 'JMS', 'Bern')
def JMS_to_Bern_wet3(C, scale, parameters):
    return wet.JMS_to_Bern(C, scale, parameters)


@wcxf.translator('WET', 'Bern', 'JMS')
def Bern_to_JMS(C, scale, parameters):
    return wet.Bern_to_JMS(C, scale, parameters)

@wcxf.translator('WET-4', 'Bern', 'JMS')
def Bern_to_JMS_wet4(C, scale, parameters):
    return wet.Bern_to_JMS(C, scale, parameters)

@wcxf.translator('WET-3', 'Bern', 'JMS')
def Bern_to_JMS_wet3(C, scale, parameters):
    return wet.Bern_to_JMS(C, scale, parameters)


@wcxf.translator('WET', 'JMS', 'formflavor')
def JMS_to_FormFlavor(C, scale, parameters):
    return wet.JMS_to_FormFlavor(C, scale, parameters)


@wcxf.translator('WET', 'FlavorKit', 'JMS')
def FlavorKit_to_JMS(C, scale, parameters):
    return wet.FlavorKit_to_JMS(C, scale, parameters)


@wcxf.translator('WET', 'JMS', 'FlavorKit')
def JMS_to_FlavorKit(C, scale, parameters):
    return wet.JMS_to_FlavorKit(C, scale, parameters)


@wcxf.translator('WET', 'FlavorKit', 'flavio')
def FlavorKit_to_flavio(C, scale, parameters):
    C_JMS = wet.FlavorKit_to_JMS(C, scale, parameters)
    return wet.JMS_to_flavio(C_JMS, scale, parameters)



import unittest
from wilson.util.qcd import alpha_s, m_b, m_c, m_s

# All numbers compared to Mathemetica version of RunDec

delta = 1e-8
deltam = 1e-5


class TestMb(unittest.TestCase):
    def test_m_b(self):
        self.assertAlmostEqual(m_b(4.2, 50, 5),
                               3.03526,
                               delta=deltam)
        self.assertAlmostEqual(m_b(4.2, 2, 4),
                               4.98814,
                               delta=deltam)
        self.assertAlmostEqual(m_b(4.2, 1, 3),
                               6.54785,
                               delta=deltam)
        self.assertAlmostEqual(m_b(4.2, 200, 6),
                               2.69632,
                               delta=deltam)


class TestMc(unittest.TestCase):
    def test_m_c(self):
        self.assertAlmostEqual(m_c(1.2, 50, 5),
                               0.612466,
                               delta=deltam)
        self.assertAlmostEqual(m_c(1.2, 2, 4),
                               1.00652,
                               delta=deltam)
        self.assertAlmostEqual(m_c(1.2, 1, 3),
                               1.32125,
                               delta=deltam)


class TestMs(unittest.TestCase):
    def test_m_s(self):
        self.assertAlmostEqual(m_s(0.1, 1, 3),
                               0.131039,
                               delta=deltam)
        self.assertAlmostEqual(m_s(0.1, 2.5, 4),
                               0.094137,
                               delta=deltam)
        self.assertAlmostEqual(m_s(0.1, 50, 5),
                               0.0607461,
                               delta=deltam)

class TestAlphas(unittest.TestCase):
    def test_alphas_invalid(self):
        with self.assertRaises(ValueError):
            alpha_s(100, 7)
        with self.assertRaises(ValueError):
            alpha_s(100, 2)
        with self.assertRaises(ValueError):
            alpha_s(100, 7)
        with self.assertRaises(ValueError):
            alpha_s(0, 6)
        with self.assertRaises(ValueError):
            alpha_s(-1, 6)
        with self.assertRaises(TypeError):
            alpha_s("1.0", 6)

    def test_alphas_5(self):
        self.assertAlmostEqual(alpha_s(100, 5),
                               0.11686431884237730186,
                               delta=delta)
        self.assertAlmostEqual(alpha_s(10, 5),
                               0.17931693160062720703,
                               delta=delta)
        # crazy values
        self.assertAlmostEqual(alpha_s(1, 5),
                               0.40957053067188524193,
                               delta=delta)
        self.assertAlmostEqual(alpha_s(1000, 5),
                               0.087076948997751428458,
                               delta=delta)

    def test_alphas_6(self):
        self.assertAlmostEqual(alpha_s(500, 6),
                               0.095517575136454583087,
                               delta=delta)
        # crazy values
        self.assertAlmostEqual(alpha_s(50, 6),
                               0.12785358110125187370,
                               delta=delta)


    def test_alphas_4(self):
        self.assertAlmostEqual(alpha_s(3, 4),
                               0.25604161478941490576,
                               delta=delta)
        # crazy values
        self.assertAlmostEqual(alpha_s(1, 4),
                               0.46414770696020787020,
                               delta=delta)
        self.assertAlmostEqual(alpha_s(1000, 4),
                               0.082139482683335368979,
                               delta=delta)


    def test_alphas_3(self):
        self.assertAlmostEqual(alpha_s(0.9, 3),
                               0.527089,
                               delta=1e-5)
        # crazy values
        self.assertAlmostEqual(alpha_s(1000, 3),
                               0.076593079980776995496,
                               delta=delta)



import unittest
import numpy as np
import numpy.testing as npt
from wilson.run.smeft import beta
from wilson.util import smeftutil
from wilson.run.smeft.tests import test_beta

C = test_beta.C.copy()
for i in C:
    if i in smeftutil.WC_keys_2f + smeftutil.WC_keys_4f:
        # make Wilson coefficients involving fermions complex!
        C[i] = C[i] + 1j*C[i]

class TestSymm(unittest.TestCase):
    def test_keys(self):
        # check no parameter or WC was forgotten in the C_symm_keys lists
        self.assertEqual(
          set(smeftutil.C_keys),
          set([c for cs in smeftutil.C_symm_keys.values() for c in cs])
        )

    def test_symmetrize_symmetric(self):
        a = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
        npt.assert_array_equal(smeftutil.symmetrize_2(a), a)
        b = np.array([[1, 2, 3], [0, 4, 5], [0, 0, 6]])
        npt.assert_array_equal(smeftutil.symmetrize_2(b), a)

    def test_symmetrize_hermitian(self):
        a = np.array([[1, 2j, 3j], [-2j, 4, 5j], [-3j, -5j, 6]])
        npt.assert_array_equal(smeftutil.symmetrize_2(a), a)
        b = np.array([[1, 2j, 3j], [0, 4, 5j], [0, 0, 6]])
        npt.assert_array_equal(smeftutil.symmetrize_2(b), a)

    def test_symmetrize_C(self):
        C_symm = smeftutil.symmetrize(C)
        # check all keys are present
        self.assertSetEqual(set(C.keys()), set(C_symm.keys()))
        for i, v in C_symm.items():
            # check trivial cases are the same
            if i in smeftutil.C_symm_keys[0] + smeftutil.C_symm_keys[1] + smeftutil.C_symm_keys[3]:
                if smeftutil.C_keys_shape[i] == 1:
                    self.assertEqual(v, C[i])
                else:
                    npt.assert_array_equal(v, C[i])
            # check symmetric
            elif i in smeftutil.C_symm_keys[9]:
                npt.assert_array_equal(v, v.T)
            # check hermitian
            elif i in smeftutil.C_symm_keys[2]:
                npt.assert_array_equal(v, v.T.conj())
            # check 2 identical FFbar
            elif i in smeftutil.C_symm_keys[4]:
                npt.assert_array_equal(v, v.transpose((2, 3, 0, 1)))
                npt.assert_array_equal(v, v.transpose((1, 0, 3, 2)).conj())
            # check 2 independent FFbar
            elif i in smeftutil.C_symm_keys[5]:
                npt.assert_array_equal(v, v.transpose((1, 0, 3, 2)).conj())
            # check special case ee
            elif i in smeftutil.C_symm_keys[6]:
                npt.assert_array_equal(v, v.transpose((2, 3, 0, 1)))
                npt.assert_array_equal(v, v.transpose((0, 3, 2, 1)))
                npt.assert_array_equal(v, v.transpose((2, 1, 0, 3)))
            # check special case qque
            elif i in smeftutil.C_symm_keys[7]:
                npt.assert_array_equal(v, v.transpose((1, 0, 2, 3)))
            # check special case qqql
            elif i in smeftutil.C_symm_keys[8]:
                # see eq. (10) of arXiv:1405.0486
                npt.assert_array_almost_equal(v + v.transpose((1, 0, 2, 3)), v.transpose((1, 2, 0, 3)) + v.transpose((2, 1, 0, 3)), decimal=15)



"""Matcher from the SMEFT 'Warsaw up' basis to the WET JMS basis.

Based on arXiv:1709.04486."""


import numpy as np
from math import sqrt, pi
import wcxf
import wilson
from wilson.parameters import p as default_parameters
from wilson.util import smeftutil, wetutil

# Based on arXiv:1709.04486

# CONSTANTS

Nc = 3

# AUXILIARY FUNCTIONS

# Eq. (6.4)
def vT(C, p):
    GF = p['GF']
    GFx = GF - sqrt(2)/4 * ( -C['ll'][1,0,0,1] - C['ll'][0,1,1,0] + 2*C['phil3'][1,1] + 2*C['phil3'][0,0] )
    return sqrt(1/sqrt(2)/abs(GFx))

# Eq. (2.22)
def eps(C, p):
    return C["phiWB"] * vT(C, p)**2

# Eq. (2.24)
def g2b(C, p):
    p["m_W"] = p["m_W"]
    return 2*p["m_W"]/vT(C, p)

def g1b(C, p):
    alpha_e = p['alpha_e']
    eb = sqrt(4*pi*alpha_e)
    return eb*g2b(C, p)/sqrt(g2b(C, p)**2-eb**2) + eb**2*g2b(C, p)/(g2b(C, p)**2-eb**2) * eps(C, p)

# Eq. (2.23)
def sb(C, p):
    return g1b(C, p)/sqrt(g1b(C, p)**2+g2b(C, p)**2) * (1 + eps(C, p)/2. * g2b(C, p)/g1b(C, p) * ((g2b(C, p)**2-g1b(C, p)**2)/(g1b(C, p)**2+g2b(C, p)**2)))
def cb(C, p):
    return g2b(C, p)/sqrt(g1b(C, p)**2+g2b(C, p)**2) * (1 - eps(C, p)/2. * g1b(C, p)/g2b(C, p) * ((g2b(C, p)**2-g1b(C, p)**2)/(g1b(C, p)**2+g2b(C, p)**2)))

# Eq. (2.26)
def eb(C, p):
    return g2b(C, p)*sb(C, p) - 1/2.*cb(C, p)*g2b(C, p)* vT(C, p)**2*C["phiWB"]
def gzb(C, p):
    return eb(C, p)/(sb(C, p)*cb(C, p)) * (1 + (g1b(C, p)**2+g2b(C, p)**2)/(2*g1b(C, p)*g2b(C, p))*vT(C, p)**2*C["phiWB"])

# Eq. (2.30)
def wl(C, p):
    return np.eye(3)+vT(C, p)**2*C["phil3"]
def wq(C, p):
    return np.eye(3)+vT(C, p)**2*C["phiq3"]
def wr(C, p):
    return 1/2.*vT(C, p)**2*C["phiud"]
def znu(C, p):
    return np.eye(3)*1/2.-1/2.*vT(C, p)**2*C["phil1"]+1/2.*vT(C, p)**2*C["phil3"]
def zel(C, p):
    return np.eye(3)*(-1/2.+sb(C, p)**2)-1/2.*vT(C, p)**2*C["phil1"]-1/2.*vT(C, p)**2*C["phil3"]
def zer(C, p):
    return np.eye(3)*sb(C, p)**2-1/2.*vT(C, p)**2*C["phie"]
def zul(C, p):
    return np.eye(3)*(1/2.-2./3*sb(C, p)**2)-1/2.*vT(C, p)**2*C["phiq1"]+1/2.*vT(C, p)**2*C["phiq3"]
def zur(C, p):
    return np.eye(3)*(-2./3)*sb(C, p)**2-1/2.*vT(C, p)**2*C["phiu"]
def zdl(C, p):
    return np.eye(3)*(-1/2.+1/3.*sb(C, p)**2)-1/2.*vT(C, p)**2*C["phiq1"]-1/2.*vT(C, p)**2*C["phiq3"]
def zdr(C, p):
    return np.eye(3)*(1/3.)*sb(C, p)**2-1/2.*vT(C, p)**2*C["phid"]

# MATCHING CONDITIONS

# initialize empty dict that will become a dict of functions
C = {}

# Table 9
# C["nu"] = lambda C, p: 1/2.*C["llphiphi"]* vT(C, p)**2

# Table 10
# C["nugamma"] = lambda C, p: np.zeros((3,3))

# Table 11
C["egamma"] = lambda C, p: 1/sqrt(2) * (-C["eW"] * sb(C, p) + C["eB"] * cb(C, p)) * vT(C, p)

C["ugamma"] = lambda C, p: 1/sqrt(2) * (C["uW"] * sb(C, p) + C["uB"] * cb(C, p)) * vT(C, p)
C["dgamma"] = lambda C, p: 1/sqrt(2) * (-C["dW"] * sb(C, p) + C["dB"] * cb(C, p)) * vT(C, p)
C["uG"] = lambda C, p: 1/sqrt(2) * C["uG"] * vT(C, p)
C["dG"] = lambda C, p: 1/sqrt(2) * C["dG"] * vT(C, p)

#Table 12
C["G"] = lambda C, p: C["G"]
C["Gtilde"] = lambda C, p: C["Gtilde"]

# Table 13
C["VnunuLL"] = lambda C, p: C["ll"]-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),znu(C, p))-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pt,sr',znu(C, p),znu(C, p))
C["VeeLL"] = lambda C, p: C["ll"]-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zel(C, p))-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pt,sr',zel(C, p),zel(C, p))
C["VnueLL"] = lambda C, p: C["ll"]+np.einsum('stpr',C["ll"])-g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pt,rs',wl(C, p),wl(C, p).conjugate())-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zel(C, p))

C["VnuuLL"] = lambda C, p: C["lq1"]+C["lq3"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zul(C, p))
C["VnudLL"] = lambda C, p: C["lq1"]-C["lq3"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zdl(C, p))
C["VeuLL"] = lambda C, p: C["lq1"]-C["lq3"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zul(C, p))
C["VedLL"] = lambda C, p: C["lq1"]+C["lq3"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zdl(C, p))
C["VnueduLL"] = lambda C, p: 2*C["lq3"]-g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pr,ts',wl(C, p),wq(C, p).conjugate())
# + h.c.

C["VuuLL"] = lambda C, p: C["qq1"]+C["qq3"]-gzb(C, p)**2/(2*p["m_Z"]**2)*np.einsum('pr,st',zul(C, p),zul(C, p))
C["VddLL"] = lambda C, p: C["qq1"]+C["qq3"]-gzb(C, p)**2/(2*p["m_Z"]**2)*np.einsum('pr,st',zdl(C, p),zdl(C, p))
C["V1udLL"] = lambda C, p: C["qq1"]+np.einsum('stpr',C["qq1"])-C["qq3"]-np.einsum('stpr',C["qq3"])+2/Nc*np.einsum('ptsr',C["qq3"])+2/Nc*np.einsum('srpt',C["qq3"])-g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pt,rs',wq(C, p),wq(C, p).conjugate())/Nc-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zul(C, p),zdl(C, p))
C["V8udLL"] = lambda C, p: 4*np.einsum('ptsr',C["qq3"])+4*np.einsum('srpt',C["qq3"])-g2b(C, p)**2/(p["m_W"]**2)*np.einsum('pt,rs',wq(C, p),wq(C, p).conjugate())


# Table 14
C["VeeRR"] = lambda C, p: C["ee"]-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pr,st',zer(C, p),zer(C, p))-gzb(C, p)**2/(4*p["m_Z"]**2)*np.einsum('pt,sr',zer(C, p),zer(C, p))

C["VeuRR"] = lambda C, p: C["eu"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zer(C, p),zur(C, p))
C["VedRR"] = lambda C, p: C["ed"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zer(C, p),zdr(C, p))

C["VuuRR"] = lambda C, p: C["uu"]-gzb(C, p)**2/(2*p["m_Z"]**2)*np.einsum('pr,st',zur(C, p),zur(C, p))
C["VddRR"] = lambda C, p: C["dd"]-gzb(C, p)**2/(2*p["m_Z"]**2)*np.einsum('pr,st',zdr(C, p),zdr(C, p))
C["V1udRR"] = lambda C, p: C["ud1"]-g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pt,rs',wr(C, p),wr(C, p).conjugate())/Nc-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zur(C, p),zdr(C, p))
C["V8udRR"] = lambda C, p: C["ud8"]-g2b(C, p)**2/(p["m_W"]**2)*np.einsum('pt,rs',wr(C, p),wr(C, p).conjugate())


# Table 15
C["VnueLR"] = lambda C, p: C["le"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zer(C, p))
C["VeeLR"] = lambda C, p: C["le"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zer(C, p))

C["VnuuLR"] = lambda C, p: C["lu"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zur(C, p))
C["VnudLR"] = lambda C, p: C["ld"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',znu(C, p),zdr(C, p))
C["VeuLR"] = lambda C, p: C["lu"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zur(C, p))
C["VedLR"] = lambda C, p: C["ld"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zel(C, p),zdr(C, p))
C["VueLR"] = lambda C, p: C["qe"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zul(C, p),zer(C, p))
C["VdeLR"] = lambda C, p: C["qe"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zdl(C, p),zer(C, p))
C["VnueduLR"] = lambda C, p: -g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pr,ts',wl(C, p),wr(C, p).conjugate())
#+ h.c.

C["V1uuLR"] = lambda C, p: C["qu1"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zul(C, p),zur(C, p))
C["V8uuLR"] = lambda C, p: C["qu8"]
C["V1udLR"] = lambda C, p: C["qd1"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zul(C, p),zdr(C, p))
C["V8udLR"] = lambda C, p: C["qd8"]
C["V1duLR"] = lambda C, p: C["qu1"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zdl(C, p),zur(C, p))
C["V8duLR"] = lambda C, p: C["qu8"]
C["V1ddLR"] = lambda C, p: C["qd1"]-gzb(C, p)**2/(p["m_Z"]**2)*np.einsum('pr,st',zdl(C, p),zdr(C, p))
C["V8ddLR"] = lambda C, p: C["qd8"]
C["V1udduLR"] = lambda C, p: -g2b(C, p)**2/(2*p["m_W"]**2)*np.einsum('pr,ts',wq(C, p),wr(C, p).conjugate())
C["V8udduLR"] = lambda C, p: np.zeros((3,3,3,3))


# Table 16
C["SeuRL"] = lambda C, p: np.zeros((3,3,3,3))
C["SedRL"] = lambda C, p: C["ledq"]
C["SnueduRL"] = lambda C, p: C["ledq"]


# Table 17
C["SeeRR"] = lambda C, p: np.zeros((3,3,3,3))

C["SeuRR"] = lambda C, p: -C["lequ1"]
C["TeuRR"] = lambda C, p: -C["lequ3"]
C["SedRR"] = lambda C, p: np.zeros((3,3,3,3))
C["TedRR"] = lambda C, p: np.zeros((3,3,3,3))
C["SnueduRR"] = lambda C, p: C["lequ1"]
C["TnueduRR"] = lambda C, p: C["lequ3"]

C["S1uuRR"] = lambda C, p: np.zeros((3,3,3,3))
C["S8uuRR"] = lambda C, p: np.zeros((3,3,3,3))
C["S1udRR"] = lambda C, p: C["quqd1"]
C["S8udRR"] = lambda C, p: C["quqd8"]
C["S1ddRR"] = lambda C, p: np.zeros((3,3,3,3))
C["S8ddRR"] = lambda C, p: np.zeros((3,3,3,3))
C["S1udduRR"] = lambda C, p: -np.einsum('stpr',C["quqd1"])
C["S8udduRR"] = lambda C, p: -np.einsum('stpr',C["quqd8"])

# # Table 18
# C["SnunuLL"] = lambda C, p: np.zeros((3,3,3,3))
#
# # Table 19
# C["SnueLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["TnueLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnueLR"] = lambda C, p: np.zeros((3,3,3,3))
#
# C["SnuuLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["TnuuLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnuuLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnudLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["TnudLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnudLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnueduLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["TnueduLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SnueduLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["VnueduRL"] = lambda C, p: np.zeros((3,3,3,3))
# C["VnueduRR"] = lambda C, p: np.zeros((3,3,3,3))

# Table 20
# C["SuddLL"] = lambda C, p: -C["qqql"]-np.einsum('rpst',C["qqql"])
# C["SduuLL"] = lambda C, p: -C["qqql"]-np.einsum('rpst',C["qqql"])
# C["SuudLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SduuLR"] = lambda C, p: -C["qque"]-np.einsum('rpst',C["qque"])
# C["SuudRL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SduuRL"] = lambda C, p: C["duql"]
# C["SdudRL"] = lambda C, p: -C["duql"]
# C["SdduRL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SduuRR"] = lambda C, p: C["duue"]
#
# # Table 21
# C["SdddLL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SuddLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SdduLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SdddLR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SdddRL"] = lambda C, p: np.zeros((3,3,3,3))
# C["SuddRR"] = lambda C, p: np.zeros((3,3,3,3))
# C["SdddRR"] = lambda C, p: np.zeros((3,3,3,3))

def match_all_array(C_SMEFT, p):
    # generate a dictionary with 0 Wilson coefficients = Standard Model
    C_SMEFT_0 = {k: 0*v for k, v in C_SMEFT.items()}
    # compute the SMEFT matching contribution but subtract the SM part
    return {k: f(C_SMEFT, p) - f(C_SMEFT_0, p) for k, f in C.items()}

def match_all(d_SMEFT, parameters=None):
    """Match the SMEFT Warsaw basis onto the WET JMS basis."""
    p = default_parameters.copy()
    if parameters is not None:
        # if parameters are passed in, overwrite the default values
        p.update(parameters)
    C = wilson.translate.smeft.wcxf2arrays(d_SMEFT)
    C = smeftutil.symmetrize(C)
    C = smeftutil.scale_dict(C)
    C = smeftutil.add_missing(C)
    C['vT'] = 246.22
    C_WET = match_all_array(C, p)
    C_WET = wilson.translate.wet.rotate_down(C_WET, p)
    C_WET = wetutil.unscale_dict_wet(C_WET)
    d_WET = wilson.translate.smeft.arrays2wcxf(C_WET)
    basis = wcxf.Basis['WET', 'JMS']
    keys = set(d_WET.keys()) & set(basis.all_wcs)
    d_WET = {k: d_WET[k] for k in keys}
    return d_WET



import unittest
import wcxf
import wilson
import numpy as np
import pkgutil


np.random.seed(235)


def get_random_wc(eft, basis, scale, cmax=1e-6):
    """Generate a random Wilson coefficient instance for a given basis."""
    basis_obj = wcxf.Basis[eft, basis]
    _wc = {}
    for s in basis_obj.sectors.values():
        for name, d in s.items():
            _wc[name] = cmax * np.random.rand()
            if 'real' not in d or not d['real']:
                _wc[name] += 1j * cmax * np.random.rand()
    return wcxf.WC(eft, basis, scale, wcxf.WC.dict2values(_wc))


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        s for c in cls.__subclasses__() for s in all_subclasses(c))


par = {
    'm_Z': 91.1876,
    'm_b': 4.18,
    'm_d': 4.8e-3,
    'm_s': 0.095,
    'm_t': 173.3,
    'm_c': 1.27,
    'm_u': 2.3e-3,
    'alpha_e': 1/127.944,
    'alpha_s': 0.1185,
    'GF': 1.1663787e-5,
}


class TestWilson(unittest.TestCase):
    def test_class(self):
        wilson.Wilson({'qd1_1123': 1}, 1000, 'SMEFT', 'Warsaw')

    def test_from_wc(self):
        wc = wcxf.WC('SMEFT', 'Warsaw', 1000, {'qd1_1123': 1})
        w1 = wilson.Wilson.from_wc(wc)
        w2 = wilson.Wilson({'qd1_1123': 1}, 1000, 'SMEFT', 'Warsaw')
        self.assertDictEqual(w1.wc.dict, w2.wc.dict)

    def test_set_initial_wcxf(self):
        test_file = pkgutil.get_data('wilson', 'data/tests/wcxf-flavio-example.yml')
        wcxf_wc = wcxf.WC.load(test_file.decode('utf-8'))
        wcxf_wc.validate()
        wilson_wc = wilson.Wilson.from_wc(wcxf_wc)
        wc_out = wilson_wc.match_run(160, 'WET', 'flavio', 'sb')
        self.assertEqual(wc_out.dict['C9_bsee'], -1+0.01j)
        self.assertEqual(wc_out.dict['C9p_bsee'], 0.1)
        self.assertEqual(wc_out.dict['C10_bsee'], 0.05j)
        self.assertEqual(wc_out.dict['C10p_bsee'], 0.1-0.3j)
        self.assertEqual(wc_out.dict.get('CS_bsee', 0), 0)

    def test_load_initial(self):
        test_file = pkgutil.get_data('wilson', 'data/tests/wcxf-flavio-example.yml')
        wilson_wc = wilson.Wilson.load_wc(test_file.decode('utf-8'))
        wc_out = wilson_wc.match_run(160, 'WET', 'flavio', 'sb')
        self.assertEqual(wc_out.dict['C9_bsee'], -1+0.01j)
        self.assertEqual(wc_out.dict['C9p_bsee'], 0.1)
        self.assertEqual(wc_out.dict['C10_bsee'], 0.05j)
        self.assertEqual(wc_out.dict['C10p_bsee'], 0.1-0.3j)
        self.assertEqual(wc_out.dict.get('CS_bsee', 0), 0)

    def test_set_initial_wcxf_minimal(self):
        for eft in ['WET', 'WET-4', 'WET-3']:
            wc = wcxf.WC(eft, 'flavio', 120, {'CVLL_sdsd': {'Im': 1}})
            ww = wilson.Wilson.from_wc(wc)
            self.assertEqual(ww.match_run(120, eft, 'flavio', 'sdsd').dict['CVLL_sdsd'], 1j)
            wc = wcxf.WC(eft, 'JMS', 120, {'VddLL_1212': {'Im': 1}})
            wc.validate()
            ww = wilson.Wilson.from_wc(wc)
            self.assertAlmostEqual(ww.match_run(120, eft, 'flavio', 'sdsd').dict['CVLL_sdsd'], 1j)

    def tets_repr(self):
        wc = wilson.Wilson.from_wc(wc)
        wc._repr_markdown_()
        wc.set_initial({'C7_bs': -0.1}, 5)
        wc._repr_markdown_()

    def test_run_wcxf(self):
        for eft in [('WET', 'WET', 120, 120), ('WET', 'WET-4', 120, 3), ('WET', 'WET-3', 120, 2),
                    ('WET-4', 'WET-4', 3, 2), ('WET-4', 'WET-3', 3, 2),
                    ('WET-3', 'WET-3', 2, 1), ]:
            wc = wcxf.WC(eft[0], 'flavio', eft[2], {'CVLL_sdsd': {'Im': 1}})
            ww = wilson.Wilson.from_wc(wc)
            wc_out = ww.match_run(eft[3], eft[1], 'flavio')
            wc_out.validate()

    def test_run_smeft(self):
        w = wilson.Wilson({'qd1_1123': 1}, 1000, 'SMEFT', 'Warsaw')
        wc = w.match_run(160, 'SMEFT', 'Warsaw up')
        wc.validate()

class TestRGsolution(unittest.TestCase):
    def test_rgsolution_smeft(self):
        wc = wcxf.WC('SMEFT', 'Warsaw', 1e5, {'qu1_1233': 1e-7})
        wc.validate()
        smeft = wilson.run.smeft.SMEFT(wc)
        sol = smeft.run_continuous(160)
        x, y = sol.plotdata('qu1_1233')
        self.assertTupleEqual(x.shape, (50,))
        self.assertTupleEqual(y.shape, (50,))
        self.assertEqual(x.dtype, float)
        self.assertEqual(y.dtype, float)

class TestWilsonConfig(unittest.TestCase):
    def test_schema(self):
        for subclass in all_subclasses(wilson.classes.ConfigurableClass):
            # check that all options in schema have a default option
            self.assertEqual(set(subclass._option_schema.schema.keys()),
                             set(subclass._default_options.keys()))
            # check that default options pass validation
            subclass._option_schema(subclass._default_options)

    def test_config(self):
        wilson.Wilson._default_options['smeft_accuracy'] = 'leadinglog'
        w = wilson.Wilson({'qd1_1123': 1}, 1000, 'SMEFT', 'Warsaw')
        self.assertEqual(w.get_option('smeft_accuracy'), 'leadinglog')
        wilson.Wilson.set_default_option('smeft_accuracy', 'integrate')
        self.assertEqual(w.get_option('smeft_accuracy'), 'leadinglog')  # not changed!
        w2 = wilson.Wilson({'qd1_1123': 1}, 1000, 'SMEFT', 'Warsaw')
        self.assertEqual(w2.get_option('smeft_accuracy'), 'integrate')  # changed!
        w.set_option('smeft_accuracy', 'leadinglog')
        self.assertEqual(w.get_option('smeft_accuracy'), 'leadinglog')
        with self.assertRaises(KeyError):
            w.get_option('my_config_doesntexist')

    def test_clearcache(self):
        w = wilson.Wilson({'CVLL_sdsd': 1}, 160, 'WET', 'flavio')
        # after init, cache empty
        self.assertDictEqual(w._cache, {})
        # run
        w.match_run(140, 'WET', 'flavio')
        # now cache not empty
        self.assertIsInstance(w._cache['WET'][140]['flavio']['all'], wcxf.WC)
        w.clear_cache()
        # now cache empty again
        self.assertDictEqual(w._cache, {})
        # check that setting option empties cache
        wilson.Wilson._default_options['smeft_accuracy'] = 666
        w = wilson.Wilson({'CVLL_sdsd': 1}, 160, 'WET', 'flavio')
        w.match_run(140, 'WET', 'flavio')
        w.set_option('smeft_accuracy', 'leadinglog')
        self.assertDictEqual(w._cache, {})



"""Utility functions and dictionaries useful for the manipulation of WET Wilson coefficients.
"""

import wcxf
import numpy as np
import ckmutil
from wilson.util.smeftutil import _d_4, _d_6


# names of Wilson coefficients with the same fermionic symmetry properties
# numbering is inspired by the corresponding categorization in SMEFT
C_symm_keys = {}
# 0 0F scalar object
C_symm_keys[0] = ["G", "Gtilde"]
# 1 2F general 3x3 matrix
C_symm_keys[1] = ["egamma", "uG","dG", "ugamma", "dgamma"]
# 3 4F general 3x3x3x3 object
C_symm_keys[3] = ['S1udRR', 'S1udduRR', 'S8udRR', 'S8udduRR', 'SedRL',
'SedRR', 'SeuRL', 'SeuRR', 'SnueduRL', 'SnueduRR', 'TedRR', 'TeuRR',
'TnueduRR', 'V1udduLR', 'V8udduLR', 'VnueduLL', 'VnueduLR']
# 4 4F two identical ffbar currents
# hermitian currents
C_symm_keys[4] = ['VuuRR', 'VddRR', 'VuuLL', 'VddLL']
# non-hermitian currents
C_symm_keys[41] = ['S1ddRR', 'S1uuRR', 'S8uuRR', 'SeeRR', 'S8ddRR']
# 5 4F two independent ffbar currents, hermitian
C_symm_keys[5] = ["VnueLL", "VnuuLL", "VnudLL", "VeuLL", "VedLL", "V1udLL",
        "V8udLL", "VeuRR", "VedRR", "V1udRR", "V8udRR", "VnueLR",
        "VeeLR", "VnuuLR", "VnudLR", "VeuLR", "VedLR", "VueLR", "VdeLR",
        "V1uuLR", "V8uuLR", "V1udLR", "V8udLR", "V1duLR", "V8duLR",
        "V1ddLR", "V8ddLR"]
# 6 4F two identical ffbar currents + Fierz symmetry
C_symm_keys[6] = ['VeeLL', 'VeeRR', 'VnunuLL']


# symmetrize JMS basis

def _scalar2array(d):
    """Convert a dictionary with scalar elements and string indices '_1234'
    to a dictionary of arrays. Unspecified entries are np.nan."""
    da = {}
    for k, v in d.items():
        if '_' not in k:
            da[k] = v
        else:
            name = ''.join(k.split('_')[:-1])
            ind = k.split('_')[-1]
            dim = len(ind)
            if name not in da:
                shape = tuple(3 for i in range(dim))
                da[name] = np.empty(shape, dtype=complex)
                da[name][:] = np.nan
            da[name][tuple(int(i) - 1 for i in ind)] = v
    return da


def _symm_herm(C):
    """To get rid of NaNs produced by _scalar2array, symmetrize operators
    where C_ijkl = C_jilk*"""
    nans = np.isnan(C)
    C[nans] = np.einsum('jilk', C)[nans].conj()
    return C


def _symm_current(C):
    """To get rid of NaNs produced by _scalar2array, symmetrize operators
    where C_ijkl = C_klij"""
    nans = np.isnan(C)
    C[nans] = np.einsum('klij', C)[nans]
    return C


def JMS_to_array(C):
    """For a dictionary with JMS Wilson coefficients, return a dictionary
    of arrays."""
    wc_keys = wcxf.Basis['WET', 'JMS'].all_wcs
    # fill in zeros for missing coefficients
    C_complete = {k: C.get(k, 0) for k in wc_keys}
    Ca = _scalar2array(C_complete)
    for k in Ca:
        if k in C_symm_keys[5]:
            Ca[k] = _symm_herm(Ca[k])
        if k in C_symm_keys[41]:
            Ca[k] = _symm_current(Ca[k])
        if k in C_symm_keys[4]:
            Ca[k] = _symm_herm(_symm_current(Ca[k]))
    return Ca


def symmetrize_JMS_dict(C):
    """For a dictionary with JMS Wilson coefficients but keys that might not be
    in the non-redundant basis, return a dictionary with keys from the basis
    and values conjugated if necessary."""
    wc_keys = set(wcxf.Basis['WET', 'JMS'].all_wcs)
    Cs = {}
    for op, v in C.items():
        if '_' not in op or op in wc_keys:
            Cs[op] = v
            continue
        name, ind = op.split('_')
        if name in C_symm_keys[5]:
            i, j, k, l = ind
            indnew = ''.join([j, i, l, k])
            Cs['_'.join([name, indnew])] = v.conjugate()
        elif name in C_symm_keys[41]:
            i, j, k, l = ind
            indnew = ''.join([k, l, i, j])
            Cs['_'.join([name, indnew])] = v
        elif name in C_symm_keys[4]:
            i, j, k, l = ind
            indnew = ''.join([l, k, j, i])
            newname = '_'.join([name, indnew])
            if newname in wc_keys:
                Cs[newname] = v.conjugate()
            else:
                indnew = ''.join([j, i, l, k])
                newname = '_'.join([name, indnew])
                if newname in wc_keys:
                    Cs[newname] = v.conjugate()
                else:
                    indnew = ''.join([k, l, i, j])
                    newname = '_'.join([name, indnew])
                    Cs[newname] = v
    return Cs


def rotate_down(C_in, p):
    """Redefinition of all Wilson coefficients in the JMS basis when rotating
    down-type quark fields from the flavour to the mass basis.

    C_in is expected to be an array-valued dictionary containg a key
    for all Wilson coefficient matrices."""
    C = C_in.copy()
    V = ckmutil.ckm.ckm_tree(p["Vus"], p["Vub"], p["Vcb"], p["gamma"])
    UdL = V
    # type dL dL dL dL
    for k in ['VddLL']:
        C[k] = np.einsum('ia,jb,kc,ld,ijkl->abcd',
                         UdL.conj(), UdL, UdL.conj(), UdL,
                         C_in[k])
    # type X X dL dL
    for k in ['V1udLL', 'V8udLL', 'VedLL', 'VnudLL']:
        C[k] = np.einsum('kc,ld,ijkl->ijcd',
                         UdL.conj(), UdL,
                         C_in[k])
    # type dL dL X X
    for k in ['V1ddLR', 'V1duLR', 'V8ddLR', 'V8duLR', 'VdeLR']:
        C[k] = np.einsum('ia,jb,ijkl->abkl',
                         UdL.conj(), UdL,
                         C_in[k])
    # type dL X dL X
    for k in ['S1ddRR', ]:
        C[k] = np.einsum('ia,kc,ijkl->ajcl',
                         UdL.conj(), UdL.conj(),
                         C_in[k])
    # type X dL X X
    for k in ['V1udduLR', 'V8udduLR']:
        C[k] = np.einsum('jb,ijkl->ibkl',
                         UdL,
                         C_in[k])
    # type X X dL X
    for k in ['VnueduLL', 'SedRR', 'TedRR', 'SnueduRR', 'TnueduRR',
              'S1udRR',  'S8udRR', 'S1udduRR',  'S8udduRR', ]:
        C[k] = np.einsum('kc,ijkl->ijcl',
                         UdL.conj(),
                         C_in[k])
    # type X X X dL
    for k in ['SedRL', ]:
        C[k] = np.einsum('ld,ijkl->ijkd',
                         UdL,
                         C_in[k])
    return C


_scale_dict = {}
for k in C_symm_keys[0]:
    _scale_dict[k] = 1
for k in C_symm_keys[1]:
    _scale_dict[k] = np.ones((3, 3))
for k in C_symm_keys[3] + C_symm_keys[5]:
    _scale_dict[k] = np.ones((3, 3, 3, 3))
for k in C_symm_keys[4] + C_symm_keys[41]:
    _scale_dict[k] = _d_4
for k in C_symm_keys[6]:
    _scale_dict[k] = _d_6


def scale_dict_wet(C):
    """To account for the fact that arXiv:Jenkins:2017jig uses a flavour
    non-redundant basis in contrast to WCxf, symmetry factors of two have to
    be introduced in several places for operators that are symmetric
    under the interchange of two currents."""
    return {k: v / _scale_dict[k] for k, v in C.items()}


def unscale_dict_wet(C):
    """Undo the scaling applied in `scale_dict`."""
    return {k: _scale_dict[k] * v for k, v in C.items()}



"""Function to perform the RG evolution for Wilson coefficients of a given
sector."""


from wilson.run.wet.definitions import sectors, coeffs
from collections import OrderedDict
import numpy as np
from functools import lru_cache
from wilson.run.wet import adm
from math import log, sqrt, pi
import wcxf

@lru_cache(maxsize=32)
def get_permissible_wcs(classname, f):
    r"""For some classes (in particular $\Delta F=1$), only a subset of Wilson
    coefficients exist in WET-3 and WET-4. Therefore, depending on the number
    of flavours `f`, the dimensionality of the ADM has to be reduced."""
    # these are the problematic sectors
    classes = ['cu', 'db', 'sb', 'sd', 'mue', 'mutau', 'taue', 'dF0']
    if classname not in classes or f == 5:
        # for 5-flavour WET, nothing to do.
        # Neither for other classes (I, II, ...) because they exist either all
        # or not at all in WET-3 and WET-4 (they have specific flavours).
        return 'all'
    if f not in [3, 4]:
        raise ValueError("f must be 3, 4, or 5.")
    if classname == 'dF0':
        sector = 'dF=0'
    else:
        sector = classname
    perm_keys = wcxf.Basis['WET-{}'.format(f), 'JMS'].sectors[sector].keys()
    all_keys = coeffs[sector]
    return [i for i, c in enumerate(all_keys) if c in perm_keys]


# new ADM functions
@lru_cache(maxsize=32)
def admeig(classname, f, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau):
    """Compute the eigenvalues and eigenvectors for a QCD anomalous dimension
    matrix that is defined in `adm.adm_s_X` where X is the name of the sector.

    Supports memoization. Output analogous to `np.linalg.eig`."""
    args = f, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau
    A = getattr(adm, 'adm_s_' + classname)(*args)
    perm_keys = get_permissible_wcs(classname, f)
    if perm_keys != 'all':
        # remove disallowed rows & columns if necessary
        A = A[perm_keys][:, perm_keys]
    w, v = np.linalg.eig(A.T)
    return w, v


@lru_cache(maxsize=32)
def getUs(classname, eta_s, f, alpha_s, alpha_e, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau):
    """Get the QCD evolution matrix."""
    w, v = admeig(classname, f, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau)
    b0s = 11 - 2 * f / 3
    a = w / (2 * b0s)
    return v @ np.diag(eta_s**a) @ np.linalg.inv(v)


@lru_cache(maxsize=32)
def getUe(classname, eta_s, f, alpha_s, alpha_e, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau):
    """Get the QCD evolution matrix."""
    args = f, m_u, m_d, m_s, m_c, m_b, m_e, m_mu, m_tau
    A = getattr(adm, 'adm_e_' + classname)(*args)
    perm_keys = get_permissible_wcs(classname, f)
    if perm_keys != 'all':
        # remove disallowed rows & columns if necessary
        A = A[perm_keys][:, perm_keys]
    w, v = admeig(classname, *args)
    b0s = 11 - 2 * f / 3
    a = w / (2 * b0s)
    K = np.linalg.inv(v) @ A.T @ v
    for i in range(K.shape[0]):
        for j in range(K.shape[1]):
            if a[i] - a[j] != 1:
                K[i, j] *= (eta_s**(a[j] + 1) - eta_s**a[i]) / (a[i] - a[j] - 1)
            else:
                K[i, j] *= eta_s**a[i] * log(1 / eta_s)
    return -alpha_e / (2 * b0s * alpha_s) * v @ K @ np.linalg.inv(v)


qG = ['uG', 'dG']
qgamma = ['ugamma', 'dgamma']
lgamma = ['egamma', 'nugamma']
G3 = ['G', 'Gtilde']


def get_m(dipole_key):
    ind = dipole_key.split('_')[-1]
    gen = int(max(ind))
    if dipole_key[:2] == 'nu':
        return 0
    elif dipole_key[0] == 'e':
        return ['m_e', 'm_mu', 'm_tau'][gen - 1]
    elif dipole_key[0] == 'u':
        return ['m_u', 'm_c', 'm_t'][gen - 1]
    elif dipole_key[0] == 'd':
        return ['m_d', 'm_s', 'm_b'][gen - 1]


def scale_C(key, p):
    g = sqrt(4 * pi * p['alpha_s'])
    e = sqrt(4 * pi * p['alpha_e'])
    name = key.split('_')[0]
    if  name in qG:
        m = p[get_m(key)]
        return g / m
    elif  name in qgamma + lgamma:
        m = p[get_m(key)]
        return g**2 / e / m
    elif key in G3:
        return g
    else:
        return 1


def run_sector(sector, C_in, eta_s, f, p_in, p_out):
    r"""Solve the WET RGE for a specific sector.

    Parameters:

    - sector: sector of interest
    - C_in: dictionary of Wilson coefficients
    - eta_s: ratio of $\alpha_s$ at input and output scale
    - f: number of active quark flavours
    - p_in: running parameters at the input scale
    - p_out: running parameters at the output scale
    """
    Cdictout = OrderedDict()
    classname = sectors[sector]
    keylist = coeffs[sector]
    if sector == 'dF=0':
        perm_keys = get_permissible_wcs('dF0', f)
    else:
        perm_keys = get_permissible_wcs(sector, f)
    if perm_keys != 'all':
        # remove disallowed keys if necessary
        keylist = np.asarray(keylist)[perm_keys]
    C_input = np.array([C_in.get(key, 0) for key in keylist])
    if np.count_nonzero(C_input) == 0 or classname == 'inv':
        # nothing to do for SM-like WCs or RG invariant operators
        C_result = C_input
    else:
        Us = getUs(classname, eta_s, f, **p_in)
        Ue = getUe(classname, eta_s, f, **p_in)
        C_scaled = [C_input[i] * scale_C(key, p_in) for i, key in enumerate(keylist)]
        C_out = (Us + Ue) @ C_scaled
        C_result = [C_out[i] / scale_C(key, p_out) for i, key in enumerate(keylist)]
    for j in range(len(C_result)):
        Cdictout[keylist[j]] = C_result[j]
    return Cdictout



import unittest
from wilson.run import wet
import wcxf
import numpy as np
import numpy.testing as npt
from wilson.run.wet import rge

np.random.seed(112)


def get_random_wc(eft, basis, scale, cmax=1e-2):
    """Generate a random wcxf.WC instance for a given basis."""
    basis_obj = wcxf.Basis[eft, basis]
    _wc = {}
    for s in basis_obj.sectors.values():
        for name, d in s.items():
            _wc[name] = cmax * np.random.rand()
            if 'real' not in d or not d['real']:
                _wc[name] += 1j * cmax * np.random.rand()
    return wcxf.WC(eft, basis, scale, wcxf.WC.dict2values(_wc))


class TestDef(unittest.TestCase):

    def test_sectors(self):
        for sname, clist in wet.definitions.coeffs.items():
            self.assertIn(sname, wcxf.Basis['WET', 'JMS'].sectors.keys())
            allkeys = wcxf.Basis['WET', 'JMS'].sectors[sname].keys()
            # for c in clist:
                # self.assertIn(c, allkeys, msg="{} not found in {}".format(c, sname))
            self.assertSetEqual(set(clist), set(allkeys), msg="Failed for {}".format(sname))


class TestClass(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.wc = get_random_wc('WET', 'JMS', 160)
        cls.wet = wet.WETrunner(cls.wc)

    def test_init(self):
        with self.assertRaises(AssertionError):
            wet.WETrunner(0)  # argument is not a WC instance
        wcf = get_random_wc('WET', 'flavio', 160)  # wrong basis
        with self.assertRaises(AssertionError):
            wet.WETrunner(wcf)

    def test_attr(self):
        self.assertEqual(self.wet.scale_in, 160)
        self.assertEqual(self.wet.C_in, self.wc.dict)

    def test_wcxf(self):
        wc = self.wet.run(4.2)
        wc.validate()

    def test_run(self):
        C_out = self.wet.run(4.2).dict
        # assert all input WCs are present in the output
        # (not vice versa as RGE can generate them from zero)
        for k in self.wet.C_in:
            self.assertTrue(k in C_out,
                            msg='{} missing in output'.format(k))


class TestEvolutionMatrices(unittest.TestCase):
    def test_inverse_s(self):
        # check inverse of QCD evolution matrices
        args = (5, 0.12, 1/128, 0, 0, 0.1, 1.2, 4.2, 0, 0.106, 1.77)
        for c in ['I', 'II', 'III', 'IV', 'IVe', 'sb', 'db', 'sd', 'mue', 'cu', 'Vb']:
            npt.assert_array_almost_equal(rge.getUs(c, 0.123, *args),
                                          np.linalg.inv(rge.getUs(c, 1/0.123, *args),),
                                          err_msg="Failed for {}".format(c))


class TestClassWET4(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.wc = get_random_wc('WET-4', 'JMS', 4)
        cls.wet = wet.WETrunner(cls.wc)

    def test_wcxf(self):
        wc = self.wet.run(1.2)
        wc.validate()

    def test_run(self):
        C_out = self.wet.run(1.2).dict
        # assert all input WCs are present in the output
        # (not vice versa as RGE can generate them from zero)
        for k in self.wet.C_in:
            self.assertTrue(k in C_out,
                            msg='{} missing in output'.format(k))



"""Solving the SMEFT RGEs."""


from . import beta
from copy import deepcopy
from math import pi, log
from scipy.integrate import solve_ivp
from wilson.util.smeftutil import C_array2dict, C_dict2array, scale_dict, unscale_dict
from wilson.translate.smeft import arrays2wcxf
import numpy as np


def smeft_evolve_leadinglog(C_in, scale_in, scale_out, newphys=True):
    """Solve the SMEFT RGEs in the leading log approximation.

    Input C_in and output C_out are dictionaries of arrays."""
    C_out = scale_dict(deepcopy(C_in))
    b = beta.beta(C_out, newphys=newphys)
    for k, C in C_out.items():
        C_out[k] = C + b[k] / (16 * pi**2) * log(scale_out / scale_in)
    return unscale_dict(C_out)


def _smeft_evolve(C_in, scale_in, scale_out, newphys=True, **kwargs):
    """Axuliary function used in `smeft_evolve` and `smeft_evolve_continuous`"""
    def fun(t0, y):
        return beta.beta_array(C=C_array2dict(y.view(complex)),
                               newphys=newphys).view(float) / (16 * pi**2)
    y0 = C_dict2array(scale_dict(C_in)).view(float)
    sol = solve_ivp(fun=fun,
                    t_span=(log(scale_in), log(scale_out)),
                    y0=y0, **kwargs)
    return sol


def smeft_evolve(C_in, scale_in, scale_out, newphys=True, **kwargs):
    """Solve the SMEFT RGEs by numeric integration.

    Input C_in and output C_out are dictionaries of arrays."""
    sol = _smeft_evolve(C_in, scale_in, scale_out, newphys=newphys, **kwargs)
    return unscale_dict(C_array2dict(sol.y[:, -1].view(complex)))


def smeft_evolve_continuous(C_in, scale_in, scale_out, newphys=True, **kwargs):
    """Solve the SMEFT RGEs by numeric integration, returning a function that
    allows to compute an interpolated solution at arbitrary intermediate
    scales."""
    sol = _smeft_evolve(C_in, scale_in, scale_out, newphys=newphys,
                        dense_output=True, **kwargs)
    @np.vectorize
    def _rge_solution(scale):
        t = log(scale)
        y = sol.sol(t).view(complex)
        yd = unscale_dict(C_array2dict(y))
        yw = arrays2wcxf(yd)
        return yw
    def rge_solution(scale):
        # this is to return a scalar if the input is scalar
        return _rge_solution(scale)[()]
    return rge_solution



import unittest
import numpy as np
import wcxf
import wilson

np.random.seed(89)

# generate a random WC instance for the SMEFT Warsaw basis
C_Warsaw_random = {}
basis = wcxf.Basis['SMEFT', 'Warsaw']
for sector, wcs in basis.sectors.items():
    for name, d in wcs.items():
         C_Warsaw_random[name] = 1e-6*np.random.rand()
         if 'real' not in d or d['real'] == False:
             C_Warsaw_random[name] += 1j*1e-6*np.random.rand()
v_Warsaw_random = wcxf.WC.dict2values(C_Warsaw_random)
wc_Warsaw_random = wcxf.WC('SMEFT', 'Warsaw', scale=160,
                           values=v_Warsaw_random)
wc_Warsaw_minimal1 = wcxf.WC('SMEFT', 'Warsaw', scale=160,
                             values={'G': 1.2})
wc_Warsaw_minimal2 = wcxf.WC('SMEFT', 'Warsaw', scale=160,
                             values={'ed_1123':  {'Re': 1.2}})

class TestWarsawMass(unittest.TestCase):
    def test_smeft_mass(self):
        for wcW in [wc_Warsaw_random, wc_Warsaw_minimal1, wc_Warsaw_minimal2]:
            wcW = wc_Warsaw_random.translate('Warsaw mass')
            p = {'Vub': 3.6e-3}  # pass a parameter, but not all
            wcW = wc_Warsaw_random.translate('Warsaw mass', p)
            wcW.validate()
            # almost all WCs should actually stay the same
            for k, v in wc_Warsaw_random.dict.items():
                if k.split('_')[0] not in ['uphi', 'uG', 'uW', 'uB', 'llphiphi']:
                    self.assertEqual(wcW.dict[k], v,
                                     msg="Not equal for {}".format(k))
            for i in range(3):
                for j in range(3):
                    if i > j:
                        # the off-diagonal neutrino mass matrix elements
                        # must vanish in the mass basis, i.e. be absent
                        self.assertTrue('llphiphi_{}{}'.format(i+1, j+1) not in wcW.dict)


class TestWarsawUp(unittest.TestCase):
    def test_warsaw_up(self):
        for wcWdown in [wc_Warsaw_random, wc_Warsaw_minimal1, wc_Warsaw_minimal2]:
            wcW = wcWdown.translate('Warsaw up')
            wcW.validate()
            # translate back and check that nothing changed
            wc_roundtrip = wcW.translate('Warsaw')
            for k, v in wcWdown.dict.items():
                self.assertAlmostEqual(v, wc_roundtrip.dict[k], places=12,
                                       msg="Failed for {}".format(k))


class TestIO(unittest.TestCase):
    def test_arrays2wcxf(self):
        """Test the functions needed for WCxf IO."""
        wc = wc_Warsaw_random
        C_arr = wilson.translate.smeft.wcxf2arrays(wc.dict)
        C_wcxf = wilson.translate.smeft.arrays2wcxf(C_arr)
        for k, v in wc.dict.items():
            self.assertEqual(v, C_wcxf[k],
                             msg="Failed for {}".format(k))



import unittest
import numpy as np
import wcxf
import wilson

np.random.seed(321)


def get_random_wc(eft, basis, cmax=1e-2):
    """Generate a random Wilson coefficient instance for a given basis."""
    basis_obj = wcxf.Basis[eft, basis]
    _wc = {}
    for s in basis_obj.sectors.values():
        for name, d in s.items():
            _wc[name] = cmax * np.random.rand()
            if 'real' not in d or not d['real']:
                _wc[name] += 1j * cmax * np.random.rand()
    return wcxf.WC(eft, basis, 80., wcxf.WC.dict2values(_wc))


class TestWETflavio(unittest.TestCase):
    def test_wet_wet4(self):
        from_wc = get_random_wc('WET', 'flavio')
        to_wc = from_wc.match('WET-4', 'flavio')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-4', 'flavio'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))

    def test_wet4_wet3(self):
        from_wc = get_random_wc('WET-4', 'flavio')
        to_wc = from_wc.match('WET-3', 'flavio')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-3', 'flavio'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))


class TestWETBern(unittest.TestCase):
    def test_wet_wet4(self):
        from_wc = get_random_wc('WET', 'Bern')
        to_wc = from_wc.match('WET-4', 'Bern')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-4', 'Bern'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))

    def test_wet4_wet3(self):
        from_wc = get_random_wc('WET-4', 'Bern')
        to_wc = from_wc.match('WET-3', 'Bern')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-3', 'Bern'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))


class TestWETJMS(unittest.TestCase):
    def test_wet_wet4(self):
        from_wc = get_random_wc('WET', 'JMS')
        to_wc = from_wc.match('WET-4', 'JMS')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-4', 'JMS'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))

    def test_wet4_wet3(self):
        from_wc = get_random_wc('WET-4', 'JMS')
        to_wc = from_wc.match('WET-3', 'JMS')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-3', 'JMS'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))

    def test_wet3_wet2(self):
        from_wc = get_random_wc('WET-3', 'JMS')
        to_wc = from_wc.match('WET-2', 'JMS')
        to_wc.validate()
        fkeys = set(to_wc.values.keys())
        fkeys_all = set(wcxf.Basis['WET-2', 'JMS'].all_wcs)
        self.assertSetEqual(fkeys_all - fkeys, set(), msg="Missing coefficients")
        for k, v in to_wc.dict.items():
            self.assertEqual(v, from_wc.dict[k], msg="Failed for {}".format(k))



from setuptools import setup, find_packages


with open('wilson/_version.py') as f:
    exec(f.read())

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

setup(name='wilson',
      version=__version__,
      author='Jason Aebischer, David M. Straub',
      author_email='jason.aebischer@tum.de, david.straub@tum.de',
      url='https://github.com/wilson-eft/wilson',
      description='Toolkit for effective field theories beyond the Standard Model',
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      license='MIT',
      packages=find_packages(),
      package_data={
      'wilson': ['run/wet/tests/data/*',
                 'run/wet/tests/data/*',
                 'data/tests/*',
                 'data/*',
                 ],
      },
      install_requires=['scipy>=1.0', 'numpy', 'pylha>=0.2', 'pyyaml',
                        'ckmutil>=0.3', 'wcxf>=1.4.4', 'rundec>=0.5',
                        'voluptuous'],
      extras_require={
            'testing': ['nose',],
      },
    )
